/* $Header$ */
/* 
   Termsize library for XFree86 OS/2 3.x, 4.0
  
   Written by Alexander Mai
   Released into Public Domain (PD)
*/

#if 1
#undef DEBUG
#else
#define DEBUG
#endif


/* Fallbacks if all methods to detect the size from APIs fail: */
#define COLUMNS_FALLBACK   80
#define ROWS_FALLBACK      25

/* What X/2 sets as default
  (well, my xinitrc.cmd script starts an xterm session withj explicit
   geometry request to 25 rows - so the xterm inherited shell will
   always keep this value, even when changed - we can't handle 
   this special case)
 */
#define X2DEFAULT_COLUMNS  80
#define X2DEFAULT_ROWS     24


#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <termio.h>
#ifdef _REXX_SUPPORT
#define INCL_REXXSAA
#endif /* _REXX_SUPPORT */
#define INCL_VIO
#include <os2.h>

/* from xf86sup - XFree86 OS/2 support driver */
#include <pty.h>

/* Prototypes for public entries */
#include "termsize.h" 

/* local stuff */
static int  _gettermsize  (int *cols, int *rows);
static int  _settermsize  (int *cols, int *rows);
static int  _getsize_env    (int *x, int *y);
static int  _getsize_vio    (const int fd, int *x, int *y);
static int  _getsize_xterm  (const int fd, int *x, int *y);
static int _believe_env   (void);

#ifdef _REXX_SUPPORT
static void MkRxString (RXSTRING *rexxstring, const char *value);
#endif /* _REXX_SUPPORT */

static inline void DPrint(const char * fmt, ...);


/* 
       Now for the implementation: 
*/

#ifdef _BACKWARD_COMPATIBILITY
/* old public interface
   make sanity check and then call local routine */
extern int
termsize (int *cols, int *rows) {

  int rc;

  rc=gettermsize(cols, rows);
  return rc;
}
#endif /* _BACKWARD_COMPATIBILITY */


/* Public interface
   Make sanity check and then call local routine */
extern int
gettermsize (int *cols, int *rows) {

  int rc;

  if (cols && rows) {
    rc = _gettermsize(cols, rows);
  }
  else {
    rc = -1;
  }
  return rc;
}


/* Public interface
   Make sanity check and then call local routine */
extern int
settermsize (int *cols, int *rows) {

  int rc;

  if (cols && rows) {
    rc = _settermsize(cols, rows);
  }
  else {
    rc = -1;
  }
  return rc;
}


/* local working horses ... */
static int
_gettermsize (int *cols, int *rows) {

#define NO_TTY (-1)

  int rc;
  int cols_x, rows_x; /* "physical" values */
  int cols_e, rows_e; /* from environment */
  int fd = NO_TTY; /* file descriptor to be used:
                  should point to valid tty or NO_TTY, 
                  the latter indicates we should use non-direct methods */

  /*
     Check for a "tty" to make sure not to investigate in
     pipes, etc.
     I first ran into this when this function didn't work as
     simply expected when called from less which in turn
     was paging a man output.
     Somehow this special case was influencing the order in which
     we look for a valid "tty".
   */

  if (isatty(STDOUT_FILENO))
	 fd = STDOUT_FILENO;
  else if (isatty(STDERR_FILENO))
	 fd = STDERR_FILENO;
  else if (isatty(STDIN_FILENO))
	 fd = STDIN_FILENO;
  else
       fd = NO_TTY;
  DPrint("fd=%i\n", fd);

  if (fd != NO_TTY) {
    char *widstr;

    rc = _getsize_env(&cols_e, &rows_e);
    widstr = getenv("WINDOWID");
    if (widstr) {
      /* assume xterm */
       DPrint("xterm assumed!\n");
       rc = _getsize_xterm(fd, &cols_x, &rows_x);
       if ( (cols_e == X2DEFAULT_COLUMNS) && (rows_e == X2DEFAULT_ROWS) ) {
           /* don't believe the return values in this situation */
           DPrint("xterm default envvars - don't believe!\n");
           cols_e = 0;
           rows_e = 0;
        }
    }
   else {
      /* assume VIO session */
      DPrint("vio assumed!\n");
      rc = _getsize_vio(fd, &cols_x, &rows_x);
      }
  } /* fd != NO_TTY */ 

  /* environment has higher rank - perhaps except when a flag 
     given ?? Or another API ignoring it?? */

  if (cols_e > 0)
    *cols = cols_e;
  else if (cols_x > 0)
    *cols = cols_x;
  else
    *cols = COLUMNS_FALLBACK;
  if (rows_e > 0)
    *rows = rows_e;
  else if (rows_x > 0)
    *rows = rows_x;
  else
    *rows = ROWS_FALLBACK;

  DPrint("cols=%i, rows=%i\n", *cols, *rows);

  return 0;
}


static int
_settermsize (int *cols, int *rows)
/* does not exist yet ... */
{

  return -1;

}


/* get COLUMNS&LINES from environment */
static int _getsize_env(int *x, int *y) {

   char *s;
   int rc=0;

   s=getenv("LINES");
   if (s)
       *y=atoi(s);
   else {
       *y = 0;
       rc++;
   }
   s=getenv("COLUMNS");
   if (s)
       *x=atoi(s);
   else {
      *x = 0;
      rc++;
   }

   return rc;
}


/* Get COLUMNS&LINES for VIO sessions */
static int _getsize_vio(const int fd, int *x, int *y) {

    VIOMODEINFO ModeData;
    int rc;

    ModeData.cb = sizeof(ModeData);
/*
    rc=VioGetMode(&ModeData, (HVIO) fd);
    Doesn't work: this handle has to be 0 for non-PM apps!?
*/
    rc=VioGetMode(&ModeData, (HVIO) 0);
    if (rc) {
      *x = 0;
      *y = 0;
      return 1;
    } else {
      *x = (int)ModeData.col;
      *y = (int)ModeData.row;
      return 0;
    }
}


/* Get COLUMNS&LINES for xterm sessions */
static int _getsize_xterm(const int fd, int *x, int *y) {
 
  int arc;
  struct winsize win;

  arc = ptioctl (fd, TIOCGWINSZ, &win);
  if (arc) {
    *x = 0;
    *y = 0;
    return 1;
  } else {
    *x=win.ws_col;
    *y=win.ws_row;
    return 0;
   }
}


 /* Believe COLUMNS, ROWS environment variables.
    Use an additional envvar for that purpose */
static int _believe_env(void) {
  const char *termsize_env;

  termsize_env = getenv("TERMSIZE_ENV");
  if (termsize_env) {
     return 1;
  }
  else
     return 0;
}


/* Get COLUMNS&LINES for VIO sessions */
static int _setsize_vio(const int fd, int *x, int *y) {

    VIOMODEINFO ModeData;
    int rc;

    ModeData.cb = sizeof(ModeData);
    ModeData.col = *x;
    ModeData.row = *y;
   
/*
    rc=VioSetMode(&ModeData, (HVIO) fd);
    Doesn't work: this handle has to be 0 for non-PM apps!?
*/
    rc=VioSetMode(&ModeData, (HVIO) 0);
    return rc;
}

#ifdef _REXX_SUPPORT
#ifdef _BACKWARD_COMPATIBILITY
extern ULONG 
rxtermsize(PSZ funcname, LONG  Argc, RXSTRING  Argv[], PSZ Queuename,
           PRXSTRING Retstr)  {

  int rc;

  rc=rxgettermsize(funcname, Argc, Argv, Queuename, Retstr);
  return rc;
}
#endif /* _BACKWARD_COMPATIBILITY */


extern ULONG 
rxgettermsize(PSZ funcname, LONG  Argc, RXSTRING  Argv[], PSZ Queuename,
           PRXSTRING Retstr)  {

  char astr[256];
  int rows, cols;

  /* get the data first */
  _gettermsize(&cols, &rows); 

  /* now do the REXX-related business */
  snprintf(astr, sizeof(astr), "%d %d", cols, rows);
  MkRxString(Retstr, astr);
  return 0;
}


extern ULONG 
rxsettermsize(PSZ funcname, LONG  Argc, RXSTRING  Argv[], PSZ Queuename,
           PRXSTRING Retstr)  {

  char astr[256];
  int rows, cols;

  /* get the data first */
  _settermsize(&cols, &rows); 

  /* now do the REXX-related business */
  snprintf(astr, sizeof(astr), "%d %d", cols, rows);
  MkRxString(Retstr, astr);
  return 0;
}


static void
MkRxString (RXSTRING *rexxstring, const char *value) {
  /* creates a valid RXSTRING from given C string.
     We don't check for any problems ... */

  if (value == NULL) {
     strcpy(rexxstring->strptr, "");
     rexxstring->strlength = 0;
  } else {
     strcpy(rexxstring->strptr, value);
     rexxstring->strlength = strlen(value);
  }
 return;
}
#endif /* _REXX_SUPPORT */


static inline void
DPrint(const char * fmt, ...) {

#ifdef DEBUG
  va_list arg_ptr;

  va_start(arg_ptr, fmt);
  vfprintf(stderr, fmt, arg_ptr);
  va_end(arg_ptr);
#endif
  return;
}
